import sys
import argparse

from glotter.run import run
from glotter.test import test
from glotter.download import download
from glotter.report import report


def main():
    parser = argparse.ArgumentParser(
        prog="glotter",
        usage="""usage: glotter [-h] COMMAND

Commands:
  run         Run sources or group of sources. Use `glotter run --help` for more information.
  test        Run tests for sources or a group of sources. Use `glotter test --help` for more information.
  download    Download all the docker images required to run the tests
  report      Output a report of discovered sources for configured projects and languages
""",
    )
    parser.add_argument(
        "command",
        type=str,
        help="Subcommand to run",
        choices=["run", "test", "download", "report"],
    )
    args = parser.parse_args(sys.argv[1:2])
    commands = {
        "download": parse_download,
        "run": parse_run,
        "test": parse_test,
        "report": parse_report,
    }
    commands[args.command]()


def parse_download():
    parser = argparse.ArgumentParser(
        prog="glotter",
        description="Download images for a source or a group of sources. This command can be filtered by language, "
        "project, or a single source. Only one option may be specified.",
    )
    _add_parallel_arg(parser, "Download images in parallel")
    args = _parse_args_for_verb(parser)
    download(args)


def parse_run():
    parser = argparse.ArgumentParser(
        prog="glotter",
        description="Run a source or a group of sources. This command can be filtered by language, project"
        "or a single source. Only one option may be specified.",
    )
    args = _parse_args_for_verb(parser)
    run(args)


def parse_test():
    parser = argparse.ArgumentParser(
        prog="glotter",
        description="Test a source or a group of sources. This command can be filtered by language, project"
        "or a single source. Only one option may be specified.",
    )
    _add_parallel_arg(parser, "Run tests in parallel")
    args = _parse_args_for_verb(parser)
    test(args)


def _add_parallel_arg(parser, help_msg):
    parser.add_argument("--parallel", action="store_true", help=help_msg)


def _parse_args_for_verb(parser):
    parser.add_argument(
        "-s",
        "--source",
        metavar="SOURCE.EXT",
        type=str,
        help="source filename (not path) to run",
    )
    parser.add_argument(
        "-p",
        "--project",
        metavar="PROJECT",
        type=str,
        help="project to run",
    )
    parser.add_argument(
        "-l",
        "--language",
        metavar="LANGUAGE",
        type=str,
        help="language to run",
    )
    args = parser.parse_args(sys.argv[2:])
    return args


def parse_report():
    parser = argparse.ArgumentParser(
        prog="glotter",
        description="Output a report of discovered sources for configured projects and languages",
    )
    parser.add_argument(
        "-o",
        "--output",
        metavar="REPORT_PATH",
        type=str,
        help="output the report as a csv at REPORT_PATH instead of to stdout",
    )
    args = parser.parse_args(sys.argv[2:])
    report(args)


if __name__ == "__main__":
    main()



# pylint hates pydantic
# pylint: disable=E0213,E0611
from typing import Optional, List, Dict, Tuple, Union, Any, Callable, ClassVar
from functools import partial

from pydantic import BaseModel, validator, constr, conlist

from glotter.utils import quote, indent

ExpectedT = Union[str, List[str], Dict[str, str]]

TransformationT = Union[str, Dict[str, List[str]]]
TransformationScalarFuncT = Callable[[str, str], Tuple[str, str]]
TransformationDictFuncT = Callable[[List[str], str, str], Tuple[str, str]]


class AutoGenParam(BaseModel):
    """Object used to auto-generated a test parameter"""

    name: str = ""
    input: Optional[str] = None
    expected: ExpectedT

    @validator("expected")
    def validate_expected(cls, value: ExpectedT) -> ExpectedT:
        """
        Validate expected value

        :param value: Expected value
        :return: Original expected value
        :raises: :exc:`ValueError` if invalid expected value
        """

        if isinstance(value, dict):
            if not value:
                raise ValueError('Too few "expected" items')

            if len(value) > 1:
                raise ValueError('Too many "expected" items')

            key, item = tuple(*value.items())
            if key == "exec":
                if not item:
                    raise ValueError('No value for "exec" item in "expected"')
            elif key != "self":
                raise ValueError(f'Invalid key "{key}" in "expected" item')

        return value

    def get_pytest_param(self) -> str:
        """
        Get pytest parameter string

        :return: pytest parameter string if name is not empty, empty string otherwise
        """

        if not self.name:
            return ""

        input_param = self.input
        if isinstance(input_param, str):
            input_param = quote(input_param)

        expected_output = self.expected
        if isinstance(expected_output, str):
            expected_output = quote(expected_output)

        return (
            f"pytest.param({input_param}, {expected_output}, id={quote(self.name)}),\n"
        )


def _append_method_to_actual(
    method: str, actual_var: str, expected_var
) -> Tuple[str, str]:
    return f"{actual_var}.{method}()", expected_var


def _append_method_to_expected(
    method: str, actual_var: str, expected_var: str
) -> Tuple[str, str]:
    return actual_var, f"{expected_var}.{method}()"


def _remove_chars(
    values: List[str], actual_var: str, expected_var: str
) -> Tuple[str, str]:
    for value in values:
        actual_var += f'.replace({quote(value)}, "")'

    return actual_var, expected_var


def _strip_chars(
    values: List[str], actual_var: str, expected_var: str
) -> Tuple[str, str]:
    for value in values:
        actual_var += f".strip({quote(value)})"

    return actual_var, expected_var


def _unique_sort(actual_var, expected_var):
    return f"sorted(set({actual_var}))", f"sorted(set({expected_var}))"


class AutoGenTest(BaseModel):
    """Object use to auto-generated a test"""

    name: constr(min_length=1, regex="^[a-zA-Z][0-9a-zA-Z_]*$")
    requires_parameters: bool = False
    params: conlist(AutoGenParam, min_items=1)
    transformations: List[TransformationT] = []

    SCALAR_TRANSFORMATION_FUNCS: ClassVar[Dict[str, TransformationScalarFuncT]] = {
        "strip": partial(_append_method_to_actual, "strip"),
        "splitlines": partial(_append_method_to_actual, "splitlines"),
        "lower": partial(_append_method_to_actual, "lower"),
        "any_order": _unique_sort,
        "strip_expected": partial(_append_method_to_expected, "strip"),
    }
    DICT_TRANSFORMATION_FUNCS: ClassVar[Dict[str, TransformationDictFuncT]] = {
        "remove": _remove_chars,
        "strip": _strip_chars,
    }

    @validator("params", each_item=True, pre=True)
    def validate_params(
        cls, value: Dict[str, Any], values: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Validate each parameter

        :param value: Parameter to validate
        :param values: Test item
        :return: Original parameter
        :raises: :exc:`ValueError` if project requires parameters but no input, no name,
            or empty name
        """

        if values.get("requires_parameters"):
            if "input" not in value:
                raise ValueError(
                    'This project requires parameters, but "input" is not specified'
                )

            if "name" not in value:
                raise ValueError(
                    'This project requires parameters, but "name" is not specified'
                )

            if not value["name"]:
                raise ValueError(
                    'This project requires parameters, but "name" is empty'
                )

        return value

    @validator("transformations", each_item=True, pre=True)
    def validate_transformation(cls, value):
        """
        Validate each transformation

        :param value: Transformation to validate
        :return: Original value
        :raises: :exc:`ValueError` if invalid transformation
        """

        if isinstance(value, str):
            if value not in cls.SCALAR_TRANSFORMATION_FUNCS:
                raise ValueError(f'Invalid transformation "{value}"')
        elif isinstance(value, dict):
            key = str(*value)
            if key not in cls.DICT_TRANSFORMATION_FUNCS:
                raise ValueError(f'Invalid transformation "{key}"')
        else:
            raise ValueError(
                f'Invalid transformation data type "{type(value).__name__}"'
            )

        return value

    def transform_vars(self) -> Tuple[str, str]:
        """
        Transform variables using the specified transformations

        :return: Transformed actual and expected variables
        """

        actual_var = "actual"
        expected_var = "expected"
        for transfomation in self.transformations:
            if isinstance(transfomation, str):
                actual_var, expected_var = self.SCALAR_TRANSFORMATION_FUNCS[
                    transfomation
                ](actual_var, expected_var)
            else:
                key, item = tuple(*transfomation.items())
                actual_var, expected_var = self.DICT_TRANSFORMATION_FUNCS[key](
                    item, actual_var, expected_var
                )

        return actual_var, expected_var

    def get_pytest_params(self) -> str:
        """
        Get pytest parameters

        :return: pytest parameters
        """

        if not self.requires_parameters:
            return ""

        pytest_params = "".join(
            indent(param.get_pytest_param(), 8) for param in self.params
        ).strip()
        return f"""\
@pytest.mark.parametrize(
    ("in_params", "expected"),
    [
        {pytest_params}
    ]
)
"""

    def get_test_function_and_run(self, project_name_underscores: str) -> str:
        """
        Get test function and run command

        :param project_name_underscores: Project name with underscores between each word
        :return: Test function and run command
        """

        func_params = ""
        run_param = ""
        if self.requires_parameters:
            func_params = "in_params, expected, "
            run_param = "params=in_params"

        return f"""\
def test_{self.name}({func_params}{project_name_underscores}):
    actual = {project_name_underscores}.run({run_param})
"""

    def get_expected_output(self, project_name_underscores: str) -> str:
        """
        Get test code that gets the expected output

        :param project_name_underscores: Project name with underscores between each word
        :return: Test code that gets the expected output

        """

        if self.requires_parameters:
            return ""

        expected_output = self.params[0].expected
        if isinstance(expected_output, str):
            expected_output = quote(expected_output)
        elif isinstance(expected_output, dict):
            return _get_expected_file(project_name_underscores, expected_output)

        return f"expected = {expected_output}\n"

    def generate_test(self, project_name_underscores: str) -> str:
        """
        Generate test code

        :param project_name_underscores: Project name with underscores between each word
        :return: Test code
        """

        test_code = "@project_test(PROJECT_NAME)\n"
        test_code += self.get_pytest_params()
        test_code += self.get_test_function_and_run(project_name_underscores)
        test_code += indent(self.get_expected_output(project_name_underscores), 4)
        actual_var, expected_var = self.transform_vars()
        test_code += indent(
            _get_assert(actual_var, expected_var, self.params[0].expected), 4
        )
        return test_code


def _get_expected_file(
    project_name_underscores: str, expected_output: Dict[str, str]
) -> str:
    if "exec" in expected_output:
        script = quote(expected_output["exec"])
        return f"expected = {project_name_underscores}.exec({script})\n"

    return f"""\
with open({project_name_underscores}.full_path, "r", encoding="utf-8") as file:
    expected = file.read()
"""


def _get_assert(actual_var: str, expected_var: str, expected_output: ExpectedT) -> str:
    if isinstance(expected_output, list):
        return f"""\
actual_list = {actual_var}
expected_list = {expected_var}
assert len(actual_list) == len(expected_list), "Length not equal"
for index in range(len(expected_list)):
    assert actual_list[index] == expected_list[index], f"Item {{index + 1}} is not equal"
"""

    test_code = ""
    if actual_var != "actual":
        test_code += f"actual = {actual_var}\n"

    if expected_var != "expected":
        test_code += f"expected = {expected_var}\n"

    return f"{test_code}assert actual == expected\n"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['glotter']

package_data = \
{'': ['*']}

install_requires = \
['black>=23.1.0,<24.0.0',
 'docker>=6.0.1,<7.0.0',
 'jinja2>=3.1.2,<4.0.0',
 'pydantic>=1.10.6,<2.0.0',
 'pytest-xdist>=3.2.0,<4.0.0',
 'pytest>=7.2.1,<8.0.0',
 'pyyaml>=6.0,<7.0']

entry_points = \
{'console_scripts': ['glotter = glotter.__main__:main']}

setup_kwargs = {
    'name': 'glotter2',
    'version': '0.5.0a4',
    'description': 'An execution library for scripts written in any language. This is a fork of https://github.com/auroq/glotter',
    'long_description': '# Glotter2\n\n[![Makefile CI](https://github.com/rzuckerm/glotter2/actions/workflows/makefile.yml/badge.svg)](https://github.com/rzuckerm/glotter2/actions/workflows/makefile.yml)\n[![Coverage](https://rzuckerm.github.io/glotter2/badge.svg)](https://rzuckerm.github.io/glotter2/html_cov)\n[![PyPI version](https://img.shields.io/pypi/v/glotter2)](https://pypi.org/project/glotter2)\n[![Python versions](https://img.shields.io/pypi/pyversions/glotter2)](https://pypi.org/project/glotter2)\n[![Python wheel](https://img.shields.io/pypi/wheel/glotter2)](https://pypi.org/project/glotter2)\n\n[![Glotter2 logo](https://rzuckerm.github.io/glotter2/_static/glotter2_small.png)](https://rzuckerm.github.io/glotter2/)\n\n*The programming language icons were downloaded from [pngegg.com](https://www.pngegg.com/)*\n\nThis is a fork of the original [Glotter](https://github.com/auroq/glotter) repository, which\nappears to be unmaintained.\n\nGlotter2 is an execution library for collections of single file scripts. It uses Docker to be able to build, run, and optionally test scripts in any language without having to install a local sdk or development environment.\n\nFor getting started with Glotter, refer to our [documentation](https://rzuckerm.github.io/glotter2/).\n\n## Contributing\n\nIf you\'d like to contribute to Glotter2, read our [contributing guidelines](./CONTRIBUTING.md).\n\n## Changelog\n\n### Glotter2 releases\n\n* 0.5.0: wip\n  * Add test generation\n  * Add `pydantic` dependency\n* 0.4.5:\n  * Add link to documentation\n* 0.4.4:\n  * Fix bug that would indicate "No tests were found" when filtering tests\n* 0.4.2:\n  * Remove call to `time.sleep` when pulling image\n* 0.4.1:\n  * Bump version since wrong version pushed to pypi\n* 0.4.0:\n  * Change test ID from `<filename>` to `<language>/<filename>`\n  * Speed up test collection by about 1 min and total test time by about\n    5 min in [sample-programs][sample-programs] by caching list of sources\n  * Modify `download`, `run`, and `test` commands so that `-p`, `-l`, and\n    `-s` are no longer mutually exclusive\n  * Add `--parallel` to `download` command to parallelize image downloads\n  * Add `--parallel` to `test` command to parallelize tests\n* 0.3.0:\n  * Fix crash when running tests for [sample-programs][sample-programs]\n    with glotter 0.2.x\n  * Upgrade dependencies to latest version:\n    * `docker >=6.0.1, <7`\n    * `Jinja >=3.1.2, <4`\n    * `pytest >=7.2.1, <8`\n    * `PyYAML >=6.0, <7`\n  * Upgrade python to 3.8 or above\n\n### Original Glotter releases\n\n* 0.2.x: Add reporting verb to output discovered sources as a table in stdout or to a csv\n* 0.1.x: Initial release of working code.\n\n[sample-programs]: https://github.com/TheRenegadeCoder/sample-programs\n',
    'author': 'auroq',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/rzuckerm/glotter2',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import os
import shutil

from black import format_str, Mode

from glotter.settings import Settings

AUTO_GEN_TEST_PATH = "test/generated"


def generate_tests():
    """
    Generate tests for all projects
    """

    shutil.rmtree(AUTO_GEN_TEST_PATH, ignore_errors=True)
    settings = Settings()
    test_generators = {
        project_name: TestGenerator(project_name, project)
        for project_name, project in settings.projects.items()
    }
    test_codes = {}
    for project_name, test_generator in test_generators.items():
        test_code = test_generator.generate_tests()
        if test_code:
            test_codes[project_name] = test_code

    for project_name, test_code in test_codes.items():
        test_generators[project_name].write_tests(test_code)


class TestGenerator:
    __test__ = False  # Indicate this is not a test

    def __init__(self, project_name, project):
        self.project_name = project_name
        self.project = project
        self.long_project_name = "_".join(self.project.words)

    def generate_tests(self):
        if not self.project.tests:
            return ""

        test_code = self._get_imports() + self._get_project_fixture()
        for test_obj in self.project.tests:
            test_code += test_obj.generate_test(self.long_project_name)

        return format_str(test_code, mode=Mode())

    def _get_imports(self):
        test_code = "from glotter import project_test, project_fixture\n"
        if self.project.requires_parameters:
            test_code += "import pytest\n"

        return test_code

    def _get_project_fixture(self):
        return f"""\
PROJECT_NAME="{self.project_name}"
@project_fixture(PROJECT_NAME)
def {self.long_project_name}(request):
    request.param.build()
    yield request.param
    request.param.cleanup()
"""

    def write_tests(self, test_code):
        os.makedirs(AUTO_GEN_TEST_PATH, exist_ok=True)
        with open(
            os.path.join(AUTO_GEN_TEST_PATH, f"test_{self.long_project_name}.py"),
            "w",
            encoding="utf-8",
        ) as f:
            f.write(test_code)
